package org.globalbioticinteractions.dataset;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.FileFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eol.globi.util.InputStreamFactory;
import org.globalbioticinteractions.cache.CacheFactory;
import org.globalbioticinteractions.cache.ProvenanceLog;
import org.globalbioticinteractions.cache.CacheUtil;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.URI;
import java.util.Collection;
import java.util.Collections;
import java.util.TreeSet;
import java.util.concurrent.atomic.AtomicReference;

public class DatasetRegistryLocal implements DatasetRegistry {
    private final static Log LOG = LogFactory.getLog(DatasetRegistryLocal.class);
    private final String cacheDir;
    private final CacheFactory cacheFactory;
    private final InputStreamFactory inputStreamFactory;

    public DatasetRegistryLocal(String cacheDir, CacheFactory cacheFactory) {
        this(cacheDir, cacheFactory, inStream -> inStream);
    }

    public DatasetRegistryLocal(String cacheDir, CacheFactory cacheFactory, InputStreamFactory factory) {
        this.cacheDir = cacheDir;
        this.cacheFactory = cacheFactory;
        this.inputStreamFactory = factory;
    }

    @Override
    public Collection<String> findNamespaces() throws DatasetRegistryException {
        File directory = new File(cacheDir);
        Collection<String> namespaces = Collections.emptyList();
        if (directory.exists() && directory.isDirectory()) {
            namespaces = collectNamespaces(directory);
        } else {
            LOG.warn("Directory [" + cacheDir + "] does not exist.");
        }
        return namespaces;
    }

    private Collection<String> collectNamespaces(File directory) throws DatasetRegistryException {
        Collection<File> accessFiles = FileUtils.listFiles(directory, new FileFileFilter() {
            @Override
            public boolean accept(File file) {
                return ProvenanceLog.PROVENANCE_LOG_FILENAME.endsWith(file.getName());
            }
        }, TrueFileFilter.INSTANCE);

        Collection<String> namespaces = new TreeSet<>();
        ProvenanceLog.ProvenanceEntryListener lineListener = values -> {
            if (values.length >= 5
                    && StringUtils.equals(values[4], CacheUtil.MIME_TYPE_GLOBI)) {
                namespaces.add(values[0]);
            }
        };

        for (File accessFile : accessFiles) {
            try (FileInputStream is = new FileInputStream(accessFile)) {
                ProvenanceLog.parseProvenanceStream(is, lineListener);
            } catch (IOException e) {
                throw new DatasetRegistryException("failed to access [" + accessFile.getAbsolutePath() + "]", e);
            }
        }
        return namespaces;
    }


    private URI findLastCachedDatasetURI(String namespace) throws DatasetRegistryException {
        AtomicReference<URI> sourceURI = new AtomicReference<>();
        ProvenanceLog.ProvenanceEntryListener provenanceEntryListener = values -> {
            if (values.length > 4
                    && StringUtils.equalsIgnoreCase(StringUtils.trim(values[0]), namespace)
                    && StringUtils.equals(StringUtils.trim(values[4]), CacheUtil.MIME_TYPE_GLOBI)) {
                sourceURI.set(URI.create(values[1]));
            }

        };
        File accessFile;
        try {
            accessFile = ProvenanceLog.findProvenanceLogFile(namespace, cacheDir);
            if (accessFile.exists()) {
                ProvenanceLog.parseProvenanceStream(new FileInputStream(accessFile), provenanceEntryListener);
            }
        } catch (IOException e) {
            throw new DatasetRegistryException("issue accessing provenance log", e);
        }
        return sourceURI.get();
    }


    @Override
    public Dataset datasetFor(String namespace) throws DatasetRegistryException {
        Dataset dataset;

        final URI sourceURI = findLastCachedDatasetURI(namespace);
        dataset = sourceURI == null ? null : new DatasetFactory(new DatasetRegistry() {
            @Override
            public Collection<String> findNamespaces() throws DatasetRegistryException {
                return Collections.singletonList(namespace);
            }

            @Override
            public Dataset datasetFor(String s) throws DatasetRegistryException {
                Dataset dataset = new DatasetImpl(namespace, sourceURI, getInputStreamFactory());
                return new DatasetWithCache(dataset,
                        cacheFactory.cacheFor(dataset));
            }
        }).datasetFor(namespace);

        if (dataset == null) {
            throw new DatasetRegistryException("failed to retrieve/cache dataset in namespace [" + namespace + "]");
        }

        return dataset;
    }

    private InputStreamFactory getInputStreamFactory() {
        return inputStreamFactory;
    }


}
